package mpe;
import BayesianNetworks.*;
import java.util.*;

public class BNVariable extends Variable {
  public int numParents;
  public BNVariable[] parents;

  int numChildren;
  BNVariable[] children;

  double CPT[];

/**
 * Constructs a variable out of a JavaBayes ProbabilityVariable.
 * Parents, children and CPT are retrieved later from the JavaBayes ProbabilityFunction.
 */
public BNVariable(ProbabilityVariable pv) {
  index = pv.get_index();
  name = pv.get_name();
  domainSize = pv.get_values().length;
  numChildren = 0;
  currentValue = -1;
  evidence = false;
}

public boolean sampleVar(){
	if(currentValue != -1) return false;
	for(int j=0; j<numParents; j++){
		if(parents[j].currentValue == -1 ) return false;
	}
	
	int otherIndex=0;
	int factor = 1;
	for(int j=numParents-1; j>=0; j--){
		otherIndex += parents[j].currentValue * factor;
		factor *= parents[j].domainSize;
	}
	
	double[] probsToSampleFrom = new double[domainSize];
	double sum = 0;
	for(int val=0; val<domainSize; val++){
		probsToSampleFrom[val] = CPT[otherIndex + val*factor];
		sum += probsToSampleFrom[val];
	}
	
	double eps = 1e-3;
	if(sum < 1-eps || sum > 1+eps) throw new RuntimeException("sum of sample probs = " + sum + ", but should be 1!");

	double rand01 = BayesianNetwork.random.nextDouble() * sum; // the * sum is important b/c sum might not be exactly 1 ! rand01 = 0.9999999 is actually sampled sometimes
//	System.out.println(rand01);
	sum = 0;
	for(int val=0; val<domainSize; val++){
		sum += probsToSampleFrom[val];
		if( sum >= rand01 ){
			currentValue = val;
			break;
		}
	}
	if(currentValue == -1) throw new RuntimeException("Sampling of variable did not work, sum = " + sum + "rand01 = " + rand01);	
	return true;
}

public void sampleProbabilities(boolean sampleStructured) throws Exception{
	double structuredProbs[][] = null;
	int numberOfPossibleProbs = -1;
	if(sampleStructured){
		String[] probStrings = IO.read("bn_probs" + domainSize + ".txt");
		numberOfPossibleProbs = probStrings.length;
		structuredProbs = new double[numberOfPossibleProbs][domainSize];

		for (int i = 0; i < probStrings.length; i++) {	
			StringTokenizer tokenizer = new StringTokenizer(probStrings[i]);
			for(int val=0; val<domainSize; val++){
				structuredProbs[i][val] = Double.parseDouble(tokenizer.nextToken());
//				System.out.print(structuredProbs[i][val]);
			}
//			System.out.println();
		}
	}

	for(int otherIndex=0; otherIndex<CPT.length/domainSize; otherIndex++){
		double[] sampleProbs = new double[domainSize];
		if(sampleStructured){
			//=== Sample one of the structured probs and then just use that.
			int randIndex = BayesianNetwork.random.nextInt(numberOfPossibleProbs);
			for(int val=0; val<domainSize; val++){
				sampleProbs[val] = structuredProbs[randIndex][val];
			}
		} else{
			//=== Sample each probability independently, then normalize.
			double sum = 0;
			for(int val=0; val<domainSize; val++){
				sampleProbs[val] = BayesianNetwork.random.nextDouble();
//				System.out.println(sampleProbs[val]);
				sum += sampleProbs[val];
			}
			for(int val=0; val<domainSize; val++){
				sampleProbs[val]/=sum;
			}

			//=== Reduce number of digits after comma. 
			int intSum = 0;
			int x;
			for(int val=0; val<domainSize-1; val++){
				x = (int) Math.floor(sampleProbs[val]*100);
				sampleProbs[val] = x/100.0;
				intSum += x;
			}
			sampleProbs[domainSize-1] = (100-intSum) / 100.0;
		}
		
		for(int val=0; val<domainSize; val++){
			CPT[otherIndex + val*(CPT.length/domainSize)] = sampleProbs[val];
		}
	}
}

void setChildren(Vector vChildren) {
 	numChildren = vChildren.size(); 
	children = new BNVariable[numChildren];
  for (int i = 0; i < numChildren; i++) {
	  children[i] = (BNVariable) vChildren.elementAt(i);
  }
}
  void setParentsAndCPTFrom(ProbabilityFunction probFunc, BayesianNetwork bn){
		CPT = probFunc.get_values();
		numParents = probFunc.number_variables()-1;
		parents = new BNVariable[numParents];
		for (int i = 0; i < numParents; i++){
			parents[i] = bn.variables[probFunc.get_variable(i+1).get_index()];
		}
  }
  public String toString(){
		String result = "Variable " + index + ": " + name + ", domainSize = " + domainSize; 
		result += "\n" + numParents + " parents: [";
		for (int i = 0; i < parents.length; i++){
			result += parents[i].index;
			if( i<numParents-1 ) result += ", ";
		}
		result += "]\nCPT = [";
		for (int i = 0; i < CPT.length; i++){
			result += CPT[i];
			if( i<CPT.length-1 ) result += "  ";
		}
		result += "]\n" + numChildren + " children: [";
		for (int i = 0; i < children.length; i++){
			result += children[i].index;
			if( i<children.length-1 ) result += "  ";
		}
		result += "]\n";
		
		return result;
  }
}
