package mpe;
import BayesianNetworks.*;
import java.util.*;


public class BayesianNetwork {
	public static Random random;
	public String name;
	public BNVariable[] variables;
	public int numVars;
	public int numProbTables;

	/**
	 * Parses the file with name 'filename' and generates a BayesianNetwork from it.
	 * @param filename java.lang.String
	 */
	public BayesianNetwork(String filename) throws Exception {
		this(new BayesNet(new java.io.DataInputStream(new java.io.FileInputStream(filename))));
	}

	/**
	 * Constructs a BayesianNetwork from a JavaBayes BayesNet.
	 */
	public BayesianNetwork(BayesNet bn) {
		if(random == null) random = new Random();
		
		//=== Initialization of basic data structures.
		name = bn.get_name();
		numVars = bn.number_variables();
		variables = new BNVariable[bn.number_variables()];
		numProbTables = bn.number_probability_functions();

		//=== Create variables.
		for (int i = 0; i < numVars; i++)
			variables[i] = new BNVariable(bn.get_probability_variable(i));

		//=== Associate correct probability tables with variables.
		//=== This can be done more efficiently, here O(n^2).
		ProbabilityVariable probVar;
		ProbabilityFunction probTable = null;

		for (int i = 0; i < numVars; i++) {
			probVar = bn.get_probability_variable(i);
			for (int j = 0; j < numProbTables; j++) {
				probTable = bn.get_probability_function(j);
				if (probTable.get_variable(0) == probVar)
					break;
			}
			//=== Set the variable's parents and CPT. 
			variables[i].setParentsAndCPTFrom(probTable, this);
		}

		//=== Compute children for each variable.
		Vector children[] = new Vector[numVars];
		for (int i = 0; i < numVars; i++)
			children[i] = new Vector();

		for (int i = 0; i < numVars; i++) {
			int numParents = variables[i].numParents;
			BNVariable[] parents = variables[i].parents;
			for (int j = 0; j < numParents; j++)
				children[parents[j].index].add(variables[i]);
		}

		//=== Set children in each variable.
		for (int i = 0; i < numVars; i++)
			variables[i].setChildren(children[i]);
	}

	public void sampleProbs(boolean sampleStructured) throws Exception{
		for (int i = 0; i < variables.length; i++) {
			variables[i].sampleProbabilities(sampleStructured);
		}
	}
	
	public void sampleEvidence(int numEvidenceVars){
		
		for(int var=0; var<numVars; var++){
			variables[var].currentValue = -1;
			variables[var].evidence = false;
		}
	
		boolean allSampled = false;
		while(!allSampled){
			allSampled = true;
			for(int var=0; var<numVars; var++){
				try {
					if(variables[var].sampleVar()) allSampled = false;
				} catch (RuntimeException e) {
					if(e.getMessage().startsWith("sum of sample probs = 0.0")){
						//=== Reset the sampled evidence and try again.
						System.out.println("Bayes net ill-specified !! But going on.");
						for(var=0; var<numVars; var++){
							variables[var].currentValue = -1;
							variables[var].evidence = false;
						}
						allSampled = false;
						break;
					} else{
						throw(e);
					}
				}
			}
		}
		
		for(int i=0; i<numEvidenceVars; i++){
			int rand;
			do{
				rand = BayesianNetwork.random.nextInt(numVars);
//				System.out.println(rand);
			} while(variables[rand].evidence); 
			variables[rand].evidence = true;
		}

//		for(int var=0; var<numVars; var++){
//			System.out.println(var + " = " + variables[var].currentValue);
//		}
	}
	
	public String toSimpleString(){
		StringBuffer toResult = new StringBuffer();
		toResult.append("network " + name +"\n");
		toResult.append( numVars + "\n");
		for (int i = 0; i < variables.length; i++) {
			BNVariable var = variables[i];
			toResult.append("v " + i + " " + var.domainSize + " " + var.parents.length + " " + var.name + "\n");
		}
		
		for (int i = 0; i < variables.length; i++) {
			BNVariable var = variables[i];
			int numParents = var.parents.length;
			toResult.append("p " + i + " ");
			for (int j = 0; j < numParents; j++) {
				toResult.append( var.parents[j].index + " ");
			}
			toResult.append("\n");
			for (int j = 0; j < var.CPT.length; j++) {
				toResult.append( var.CPT[j] + " ");
			}
			toResult.append("\n");
		}
		
		toResult.append("obs\n");
		for(int var=0; var<numVars; var++){
			if(variables[var].evidence)
				toResult.append(var	+ " " + variables[var].currentValue +"\n");
		}
		return toResult.toString();
	}
	
	public String toBIFString(){
		StringBuffer toResult = new StringBuffer();
		toResult.append("network " + name + "{\n}\n");
		for (int i = 0; i < variables.length; i++) {
			BNVariable var = variables[i];
			toResult.append("variable " + var.name + "{\n");
			toResult.append("  type discrete [ " + var.domainSize + "  ] { ");
			for (int j = 0; j < var.domainSize-1; j++) {
				toResult.append( j + ", ");
			}
			toResult.append(var.domainSize-1 + " };\n}\n");
		}
		for (int i = 0; i < variables.length; i++) {
			BNVariable var = variables[i];
			int numParents = var.parents.length;
			toResult.append("probability ( " + var.name );
			if( numParents > 0 ){
				toResult.append( " | " ); 
 				for (int j = 0; j < numParents-1; j++) {
					toResult.append( var.parents[j].name + ", ");
				}
				toResult.append(var.parents[numParents-1].name );
			} 
			toResult.append( " ) {\n");
			for (int j = 0; j < var.CPT.length/var.domainSize; j++) {
				if( numParents > 0 ){
					toResult.append( "  (");
					int tmp = j;
					int[] parentVals = new int[numParents];
					for (int k = numParents-1; k >= 0; k--) {
						parentVals[k] = tmp%var.parents[k].domainSize;
						tmp /= var.parents[k].domainSize;
					} 
					for (int k = 0; k < numParents-1; k++) {
						toResult.append(parentVals[k] + ", ");
					}
					toResult.append(parentVals[numParents-1] + ") ");
				} else {
					toResult.append( "  table ");
				}
				for (int k = 0; k < var.domainSize-1; k++) {
					toResult.append( var.CPT[j+k*(var.CPT.length/var.domainSize)] + ", ");
				}
				toResult.append( var.CPT[j+(var.domainSize-1)*(var.CPT.length/var.domainSize)] + ";\n");
			}
			toResult.append( "}\n");
		}
		return toResult.toString();
	}
	
	public String toKKString() {
		String result = "//Bayesian Network \"" + name + "\"\n";
		result += "//================\n";
		result += "FORMAT3\n";
		int numLinks = 0;
		for (int i = 0; i < numVars; i++) {
			numLinks += variables[i].numParents;
		}
		result += numVars + " " + numLinks + "\n";

		for (int i = 0; i < numVars; i++) {
			result += "CT ";
			BNVariable var = variables[i];

			for (int j = 0; j < var.numParents; j++) {
				result += var.parents[j].index + " ";
			}
			result += var.index + "\n";

			int assignmentThisVar = -1;
			int[] assignmentParents = new int[var.numParents];
			for (int j = 0; j < assignmentParents.length; j++) {
				assignmentParents[j] = 0;
			}

			for (int j = 0; j < var.CPT.length; j++) {
				int tmp = j;

				assignmentThisVar++;
				if (assignmentThisVar == var.domainSize) {
					assignmentThisVar = 0;
					int index = var.numParents - 1;
					while (index >= 0) {
						assignmentParents[index]++;
						if (assignmentParents[index] == var.parents[index].domainSize) {
							assignmentParents[index] = 0;
							index--;
						} else {
							break;
						}
					}
					if (index < 0)
						System.out.println(
							"More entries in CPT than value combinations available");
				}
				for (int k = 0; k < var.numParents; k++) {
					result += assignmentParents[k] + " ";
				}
				result += assignmentThisVar + " ";
				int index = 0;
				int factor = 1;
				for (int k = assignmentParents.length-1; k >= 0; k--) {
					index += factor*assignmentParents[k];
					factor *= var.parents[k].domainSize;
				}
				index += factor*assignmentThisVar;
				factor *= var.domainSize;
//				result += "var.CPT[" + index + "]" + "\n";
				result += var.CPT[index] + "\n";
			}
		}
		return result;
	}
	public String toString() {
		String result = "Bayesian Network \"" + name + "\"\n";
		result += "================\n\n";
		result += (numVars + " variables:\n");
		for (int i = 0; i < numVars; i++) {
			result += variables[i].toString() + "\n";
		}
		result += "\n\nAnd as KK String:\n\n" + toKKString();
		return result;
	}
	

}
